/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.hadoop.hive.ql.exec;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.hive.conf.HiveConf;
import org.apache.hadoop.hive.ql.metadata.HiveException;
import org.apache.hadoop.hive.ql.plan.exprNodeColumnDesc;
import org.apache.hadoop.hive.ql.plan.exprNodeDesc;
import org.apache.hadoop.hive.ql.plan.joinCond;
import org.apache.hadoop.hive.ql.plan.joinDesc;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.hive.serde2.objectinspector.InspectableObject;
import org.apache.hadoop.hive.serde2.objectinspector.ObjectInspector;
import org.apache.hadoop.hive.serde2.objectinspector.ObjectInspectorUtils;
import org.apache.hadoop.hive.serde2.objectinspector.ObjectInspectorFactory;
import org.apache.hadoop.mapred.Reporter;

/**
 * Join operator implementation.
 */
public class JoinOperator extends Operator<joinDesc> implements Serializable {

  static final private Log LOG = LogFactory.getLog(JoinOperator.class.getName());
  
  // a list of value expressions for each alias are maintained
  public static class JoinExprMap {
    ExprNodeEvaluator[] valueFields;

    public JoinExprMap(ExprNodeEvaluator[] valueFields) {
      this.valueFields = valueFields;
    }

    public ExprNodeEvaluator[] getValueFields() {
      return valueFields;
    }

  }

  public static class IntermediateObject {
    ArrayList<Object>[] objs;
    int curSize;

    public IntermediateObject(ArrayList<Object>[] objs, int curSize) {
      this.objs = objs;
      this.curSize = curSize;
    }

    public ArrayList<Object>[] getObjs() {
      return objs;
    }

    public int getCurSize() {
      return curSize;
    }

    public void pushObj(ArrayList<Object> obj) {
      objs[curSize++] = obj;
    }

    public void popObj() {
      curSize--;
    }
  }

  transient protected int numValues; // number of aliases
  transient static protected ExprNodeEvaluator aliasField;
  transient static protected ExprNodeEvaluator keyField;
  transient protected HashMap<Byte, JoinExprMap> joinExprs;
  transient static protected Byte[] order; // order in which the results should
                                           // be outputted
  transient protected joinCond[] condn;
  transient protected boolean noOuterJoin;
  transient private Object[] dummyObj; // for outer joins, contains the
                                       // potential nulls for the concerned
                                       // aliases
  transient private Vector<ArrayList<Object>>[] dummyObjVectors;
  transient private Stack<Iterator<ArrayList<Object>>> iterators;
  transient private int totalSz; // total size of the composite object
  transient ObjectInspector joinOutputObjectInspector;

  static {
    aliasField = ExprNodeEvaluatorFactory.get(new exprNodeColumnDesc(
        String.class, Utilities.ReduceField.ALIAS.toString()));
    keyField = ExprNodeEvaluatorFactory.get(new exprNodeColumnDesc(
        String.class, Utilities.ReduceField.KEY.toString()));
  }

  HashMap<Byte, Vector<ArrayList<Object>>> storage;
  int joinEmitInterval = -1;
  
  public void initialize(Configuration hconf, Reporter reporter) throws HiveException {
    super.initialize(hconf, reporter);
    
    totalSz = 0;
    // Map that contains the rows for each alias
    storage = new HashMap<Byte, Vector<ArrayList<Object>>>();

    numValues = conf.getExprs().size();
    joinExprs = new HashMap<Byte, JoinExprMap>();
    if (order == null) {
      order = new Byte[numValues];
      for (int i = 0; i < numValues; i++)
        order[i] = (byte) i;
    }
    condn = conf.getConds();
    noOuterJoin = conf.getNoOuterJoin();
    Map<Byte, ArrayList<exprNodeDesc>> map = conf.getExprs();
    Iterator entryIter = map.entrySet().iterator();
    while (entryIter.hasNext()) {
      Map.Entry e = (Map.Entry) entryIter.next();
      Byte key = (Byte) e.getKey();
      ArrayList<exprNodeDesc> expr = (ArrayList<exprNodeDesc>) e.getValue();
      int sz = expr.size();
      totalSz += sz;

      ExprNodeEvaluator[] valueFields = new ExprNodeEvaluator[sz];

      for (int j = 0; j < sz; j++)
        valueFields[j] = ExprNodeEvaluatorFactory.get(expr.get(j));

      joinExprs.put(key, new JoinExprMap(valueFields));
    }

    ArrayList<ObjectInspector> structFieldObjectInspectors = new ArrayList<ObjectInspector>(
        totalSz);
    for (int i = 0; i < totalSz; i++) {
      structFieldObjectInspectors.add(ObjectInspectorFactory
          .getStandardPrimitiveObjectInspector(String.class));
    }
    joinOutputObjectInspector = ObjectInspectorFactory
        .getStandardStructObjectInspector(ObjectInspectorUtils
            .getIntegerArray(totalSz), structFieldObjectInspectors);

    dummyObj = new Object[numValues];
    dummyObjVectors = new Vector[numValues];

    int pos = 0;
    for (Byte alias : order) {
      int sz = map.get(alias).size();
      ArrayList<Object> nr = new ArrayList<Object>(sz);

      for (int j = 0; j < sz; j++)
        nr.add(null);

      dummyObj[pos] = nr;
      Vector<ArrayList<Object>> values = new Vector<ArrayList<Object>>();
      values.add((ArrayList<Object>) dummyObj[pos]);
      dummyObjVectors[pos] = values;
      pos++;
    }

    iterators = new Stack<Iterator<ArrayList<Object>>>();
    
    joinEmitInterval = HiveConf.getIntVar(hconf, HiveConf.ConfVars.HIVEJOINEMITINTERVAL);
  }

  public void startGroup() throws HiveException {
    LOG.trace("Join: Starting new group");
    storage.clear();
    for (Byte alias : order)
      storage.put(alias, new Vector<ArrayList<Object>>());
  }

  InspectableObject tempAliasInspectableObject = new InspectableObject();

  public void process(Object row, ObjectInspector rowInspector)
      throws HiveException {
    try {
      // get alias
      aliasField.evaluate(row, rowInspector, tempAliasInspectableObject);
      Byte alias = (Byte) (tempAliasInspectableObject.o);

      // get the expressions for that alias
      JoinExprMap exmap = joinExprs.get(alias);
      ExprNodeEvaluator[] valueFields = exmap.getValueFields();

      // Compute the values
      ArrayList<Object> nr = new ArrayList<Object>(valueFields.length);
      for (ExprNodeEvaluator vField : valueFields) {
        vField.evaluate(row, rowInspector, tempAliasInspectableObject);
        nr.add(tempAliasInspectableObject.o);
      }

      // Are we consuming too much memory
      if (storage.get(alias).size() == joinEmitInterval) {
        if (alias == numValues - 1) {
          // The input is sorted by alias, so if we are already in the last join
          // operand,
          // we can emit some results now.
          // Note this has to be done before adding the current row to the
          // storage,
          // to preserve the correctness for outer joins.
          checkAndGenObject();
          storage.get(alias).clear();
        } else {
          // Output a warning if we reached at least 1000 rows for a join
          // operand
          // We won't output a warning for the last join operand since the size
          // will never goes to joinEmitInterval.
          InspectableObject io = new InspectableObject();
          keyField.evaluate(row, rowInspector, io);
          LOG.warn("table " + alias
              + " has more than joinEmitInterval rows for join key " + io.o);
        }
      }

      // Add the value to the vector
      storage.get(alias).add(nr);

    } catch (Exception e) {
      e.printStackTrace();
      throw new HiveException(e);
    }
  }

  private void createForwardJoinObject(IntermediateObject intObj,
      boolean[] nullsArr) throws HiveException {
    ArrayList<Object> nr = new ArrayList<Object>(totalSz);
    for (int i = 0; i < numValues; i++) {
      Byte alias = order[i];
      int sz = joinExprs.get(alias).getValueFields().length;
      if (nullsArr[i]) {
        for (int j = 0; j < sz; j++) {
          nr.add(null);
        }
      } else {
        ArrayList<Object> obj = intObj.getObjs()[i];
        for (int j = 0; j < sz; j++) {
          nr.add(obj.get(j));
        }
      }
    }
    forward(nr, joinOutputObjectInspector);
  }

  private void copyOldArray(boolean[] src, boolean[] dest) {
    for (int i = 0; i < src.length; i++)
      dest[i] = src[i];
  }

  private Vector<boolean[]> joinObjectsInnerJoin(Vector<boolean[]> resNulls,
      Vector<boolean[]> inputNulls, ArrayList<Object> newObj,
      IntermediateObject intObj, int left, boolean newObjNull) {
    if (newObjNull)
      return resNulls;
    Iterator<boolean[]> nullsIter = inputNulls.iterator();
    while (nullsIter.hasNext()) {
      boolean[] oldNulls = nullsIter.next();
      boolean oldObjNull = oldNulls[left];
      if (!oldObjNull) {
        boolean[] newNulls = new boolean[intObj.getCurSize()];
        copyOldArray(oldNulls, newNulls);
        newNulls[oldNulls.length] = false;
        resNulls.add(newNulls);
      }
    }
    return resNulls;
  }

  private Vector<boolean[]> joinObjectsLeftOuterJoin(
      Vector<boolean[]> resNulls, Vector<boolean[]> inputNulls,
      ArrayList<Object> newObj, IntermediateObject intObj, int left,
      boolean newObjNull) {
    Iterator<boolean[]> nullsIter = inputNulls.iterator();
    while (nullsIter.hasNext()) {
      boolean[] oldNulls = nullsIter.next();
      boolean oldObjNull = oldNulls[left];
      boolean[] newNulls = new boolean[intObj.getCurSize()];
      copyOldArray(oldNulls, newNulls);
      if (oldObjNull)
        newNulls[oldNulls.length] = true;
      else
        newNulls[oldNulls.length] = newObjNull;
      resNulls.add(newNulls);
    }
    return resNulls;
  }

  private Vector<boolean[]> joinObjectsRightOuterJoin(
      Vector<boolean[]> resNulls, Vector<boolean[]> inputNulls,
      ArrayList<Object> newObj, IntermediateObject intObj, int left,
      boolean newObjNull, boolean firstRow) {
    if (newObjNull)
      return resNulls;

    if (inputNulls.isEmpty() && firstRow) {
      boolean[] newNulls = new boolean[intObj.getCurSize()];
      for (int i = 0; i < intObj.getCurSize() - 1; i++)
        newNulls[i] = true;
      newNulls[intObj.getCurSize()-1] = newObjNull;
      resNulls.add(newNulls);
      return resNulls;
    }

    boolean allOldObjsNull = firstRow;

    Iterator<boolean[]> nullsIter = inputNulls.iterator();
    while (nullsIter.hasNext()) {
      boolean[] oldNulls = nullsIter.next();
      if (!oldNulls[left]) {
        allOldObjsNull = false;
        break;
      }
    }

    nullsIter = inputNulls.iterator();
    while (nullsIter.hasNext()) {
      boolean[] oldNulls = nullsIter.next();
      boolean oldObjNull = oldNulls[left];

      if (!oldObjNull) {
        boolean[] newNulls = new boolean[intObj.getCurSize()];
        copyOldArray(oldNulls, newNulls);
        newNulls[oldNulls.length] = newObjNull;
        resNulls.add(newNulls);
      } else if (allOldObjsNull) {
        boolean[] newNulls = new boolean[intObj.getCurSize()];
        for (int i = 0; i < intObj.getCurSize() - 1; i++)
          newNulls[i] = true;
        newNulls[oldNulls.length] = newObjNull;
        resNulls.add(newNulls);
        return resNulls;
      }
    }
    return resNulls;
  }

  private Vector<boolean[]> joinObjectsFullOuterJoin(
      Vector<boolean[]> resNulls, Vector<boolean[]> inputNulls,
      ArrayList<Object> newObj, IntermediateObject intObj, int left,
      boolean newObjNull, boolean firstRow) {
    if (newObjNull) {
      Iterator<boolean[]> nullsIter = inputNulls.iterator();
      while (nullsIter.hasNext()) {
        boolean[] oldNulls = nullsIter.next();
        boolean[] newNulls = new boolean[intObj.getCurSize()];
        copyOldArray(oldNulls, newNulls);
        newNulls[oldNulls.length] = newObjNull;
        resNulls.add(newNulls);
      }
      return resNulls;
    }

    if (inputNulls.isEmpty() && firstRow) {
      boolean[] newNulls = new boolean[intObj.getCurSize()];
      for (int i = 0; i < intObj.getCurSize() - 1; i++)
        newNulls[i] = true;
      newNulls[intObj.getCurSize()-1] = newObjNull;
      resNulls.add(newNulls);
      return resNulls;
    }

    boolean allOldObjsNull = firstRow;

    Iterator<boolean[]> nullsIter = inputNulls.iterator();
    while (nullsIter.hasNext()) {
      boolean[] oldNulls = nullsIter.next();
      if (!oldNulls[left]) {
        allOldObjsNull = false;
        break;
      }
    }
    boolean rhsPreserved = false;

    nullsIter = inputNulls.iterator();
    while (nullsIter.hasNext()) {
      boolean[] oldNulls = nullsIter.next();
      boolean oldObjNull = oldNulls[left];

      if (!oldObjNull) {
        boolean[] newNulls = new boolean[intObj.getCurSize()];
        copyOldArray(oldNulls, newNulls);
        newNulls[oldNulls.length] = newObjNull;
        resNulls.add(newNulls);
      } else if (oldObjNull) {
        boolean[] newNulls = new boolean[intObj.getCurSize()];
        copyOldArray(oldNulls, newNulls);
        newNulls[oldNulls.length] = true;
        resNulls.add(newNulls);

        if (allOldObjsNull && !rhsPreserved) {
          newNulls = new boolean[intObj.getCurSize()];
          for (int i = 0; i < oldNulls.length; i++)
            newNulls[i] = true;
          newNulls[oldNulls.length] = false;
          resNulls.add(newNulls);
          rhsPreserved = true;
        }
      }
    }
    return resNulls;
  }

  /*
   * The new input is added to the list of existing inputs. Each entry in the
   * array of inputNulls denotes the entries in the intermediate object to be
   * used. The intermediate object is augmented with the new object, and list of
   * nulls is changed appropriately. The list will contain all non-nulls for a
   * inner join. The outer joins are processed appropriately.
   */
  private Vector<boolean[]> joinObjects(Vector<boolean[]> inputNulls,
                                        ArrayList<Object> newObj, IntermediateObject intObj, 
                                        int joinPos, boolean firstRow) {
    Vector<boolean[]> resNulls = new Vector<boolean[]>();
    boolean newObjNull = newObj == dummyObj[joinPos] ? true : false;
    if (joinPos == 0) {
      if (newObjNull)
        return null;
      boolean[] nulls = new boolean[1];
      nulls[0] = newObjNull;
      resNulls.add(nulls);
      return resNulls;
    }

    int left = condn[joinPos - 1].getLeft();
    int type = condn[joinPos - 1].getType();

    // process all nulls for RIGHT and FULL OUTER JOINS
    if (((type == joinDesc.RIGHT_OUTER_JOIN) || (type == joinDesc.FULL_OUTER_JOIN))
        && !newObjNull && (inputNulls == null) && firstRow) {
      boolean[] newNulls = new boolean[intObj.getCurSize()];
      for (int i = 0; i < newNulls.length - 1; i++)
        newNulls[i] = true;
      newNulls[newNulls.length - 1] = false;
      resNulls.add(newNulls);
      return resNulls;
    }

    if (inputNulls == null)
      return null;

    if (type == joinDesc.INNER_JOIN)
      return joinObjectsInnerJoin(resNulls, inputNulls, newObj, intObj, left,
          newObjNull);
    else if (type == joinDesc.LEFT_OUTER_JOIN)
      return joinObjectsLeftOuterJoin(resNulls, inputNulls, newObj, intObj,
          left, newObjNull);
    else if (type == joinDesc.RIGHT_OUTER_JOIN)
      return joinObjectsRightOuterJoin(resNulls, inputNulls, newObj, intObj,
                                       left, newObjNull, firstRow);
    assert (type == joinDesc.FULL_OUTER_JOIN);
    return joinObjectsFullOuterJoin(resNulls, inputNulls, newObj, intObj, left,
                                    newObjNull, firstRow);
  }

  /*
   * genObject is a recursive function. For the inputs, a array of bitvectors is
   * maintained (inputNulls) where each entry denotes whether the element is to
   * be used or not (whether it is null or not). The size of the bitvector is
   * same as the number of inputs under consideration currently. When all inputs
   * are accounted for, the output is forwared appropriately.
   */
  private void genObject(Vector<boolean[]> inputNulls, int aliasNum,
                         IntermediateObject intObj, boolean firstRow) throws HiveException {
    boolean childFirstRow = firstRow;
    if (aliasNum < numValues) {
      Iterator<ArrayList<Object>> aliasRes = storage.get(order[aliasNum])
          .iterator();
      iterators.push(aliasRes);
      while (aliasRes.hasNext()) {
        ArrayList<Object> newObj = aliasRes.next();
        intObj.pushObj(newObj);
        Vector<boolean[]> newNulls = joinObjects(inputNulls, newObj, intObj,
                                                 aliasNum, childFirstRow);
        genObject(newNulls, aliasNum + 1, intObj, firstRow);
        intObj.popObj();
        firstRow = false;
      }
      iterators.pop();
    } else {
      if (inputNulls == null)
        return;
      Iterator<boolean[]> nullsIter = inputNulls.iterator();
      while (nullsIter.hasNext()) {
        boolean[] nullsVec = nullsIter.next();
        createForwardJoinObject(intObj, nullsVec);
      }
    }
  }

  /**
   * Forward a record of join results.
   * 
   * @throws HiveException
   */
  public void endGroup() throws HiveException {
    LOG.trace("Join Op: endGroup called: numValues=" + numValues);
    checkAndGenObject();
  }

  private void checkAndGenObject() throws HiveException {
    // does any result need to be emitted
    for (int i = 0; i < numValues; i++) {
      Byte alias = order[i];
      if (storage.get(alias).iterator().hasNext() == false) {
        if (noOuterJoin) {
          LOG.trace("No data for alias=" + i);
          return;
        } else {
          storage.put(alias, dummyObjVectors[i]);
        }
      }
    }

    LOG.trace("calling genObject");
    genObject(null, 0, new IntermediateObject(new ArrayList[numValues], 0), true);
    LOG.trace("called genObject");
  }

  /**
   * All done
   * 
   */
  public void close(boolean abort) throws HiveException {
    LOG.trace("Join Op close");
    super.close(abort);
  }

}
